/* -*- Mode: javascript; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * Copyright (c) 2006, 2007 ActiveState Software Inc.
 * Refer to the license in the Komodo IDE installation
 * for the licensing information of this source code.
 *
 * winger-mediator.p.js -- manage the script/breakpoint relationship
 * ref the Mediator design pattern.
 */

if (typeof(JSDBGP) == 'undefined') {
    var JSDBGP = {};
}

// A script here can contain 1+ breakpoints.   Drop any scripts that
// don't have any breakpoints.

// Breakpoints can be associated with 0+ scripts.
// This class models that.

/* API for JSDBGP.BreakpointScriptMediator:
 * -- source-code oriented functions
 * reportBreakpointAdd(url, lineno, bkpt_id)
 * reportBreakpointDelete(url, lineno, bkpt_id)
 * getSourceActiveURLs => [list of url]
 * getSourceBreakpoints(url) => { lineNo => bkpt_id }, view-based
 * getSourceBreakpointAtLine(url, line)
 * -- tag/id/pc functions
 * reportAddSrcBkpt(tag, bkpt_id, pc)
 * reportAddPPBkpt(tag, bkpt_id, pc)
 * reportBkptDelete(bkpt_id)
 * getBoundTags(bkpt_id)
 * getBoundSrcBkpts(tag)
 * getBoundPPBkpts(tag)
 * getAnyBkpt(tag, pc)
 * getPC(tag, bkpt_id)
 * -- "broadcasters"
 * doScriptDeleted(tag)
 */

JSDBGP.BreakpointScriptMediator = function(mainObj) {
    this.mainObj = mainObj;
    this.boundBptsByTag = {};    // tag => BkptListHead
    this.breakpointsByID = {};   // bkpt_id => {tag => null}

    this.unboundBreakpointLocations = new JSDBGP.BreakpointScriptMediator.UnboundBreakpointLocations(mainObj);
    this.log = JSDBGP.Logging.getLogger("winger.breakpointScriptMediator");
    this.log.setLevel(JSDBGP.Logging.DEBUG);
};

JSDBGP.BreakpointScriptMediator.TagList = function() {
    this.by_pc = {}; // pc => bkpt_id
    this.by_id = {}; // bkpt_id => pc
}

JSDBGP.BreakpointScriptMediator.BkptListHead = function() {
    this.pp = new JSDBGP.BreakpointScriptMediator.TagList();
    this.src = new JSDBGP.BreakpointScriptMediator.TagList();
}





































// This module consists of observers (report*) and getters, and broadcasters (do*).
// Observers are told something is happening, and update data-structures.
// Getters walk data structures.
// Broadcasters figure out what to tell suboridinate modules.

JSDBGP.BreakpointScriptMediator.prototype.reportBreakpointAdd = function(url, lineno, bkpt_id) {
    this.unboundBreakpointLocations.addLoc(url, lineno, bkpt_id);
};

JSDBGP.BreakpointScriptMediator.prototype.reportBreakpointDelete = function(url, lineno, bkpt_id) {
    this.unboundBreakpointLocations.deleteLoc(url, lineno, bkpt_id);






};

JSDBGP.BreakpointScriptMediator.prototype.getSourceActiveURLs = function() {
    var res = this.unboundBreakpointLocations.getURLs();
    return res;
};

// returns a hash of {line => {bkpt_id => null}}
JSDBGP.BreakpointScriptMediator.prototype.getSourceBreakpoints = function(url) {
    var res = this.unboundBreakpointLocations.getLoc(url);
    return res;
};

// returns a hash of {bkpt_id => null}
JSDBGP.BreakpointScriptMediator.prototype.getSourceBreakpointAtLine = function(url, line) {
    var lines = this.getSourceBreakpoints(url);
    if (!lines) {
        return null;
    }
    return lines[line];
};

//////////////////////////////////////////////////////////////////
// Intern helpers
//////////////////////////////////////////////////////////////////

JSDBGP.BreakpointScriptMediator.prototype._internTag = function(tag) {
    if (!(tag in this.boundBptsByTag)) {
        this.boundBptsByTag[tag] = new JSDBGP.BreakpointScriptMediator.BkptListHead();
    };
    return this.boundBptsByTag[tag];
}

JSDBGP.BreakpointScriptMediator.prototype._internID = function(bkpt_id) {
    if (!(bkpt_id in this.breakpointsByID)) {
        this.breakpointsByID[bkpt_id] = {}; // hash of tag => null
    };
    return this.breakpointsByID[bkpt_id];
}

JSDBGP.BreakpointScriptMediator.prototype._reportBkptAdd = function(tag, bkpt_id, pc, which) {
    var hd = this._internTag(tag)[which];
    hd.by_pc[pc] = bkpt_id;
    hd.by_id[bkpt_id] = pc;
    /*
    this.log.debug("@@@@ - boundBptsByTag["
                   + tag
                   + "]."
                   + which
                   + ".by_pc["
                   + pc
                   + "] = " + bkpt_id);
    this.log.debug("@@@@ - boundBptsByTag["
                   + tag
                   + "]."
                   + which
                   + ".by_id["
                   + bkpt_id
                   + "] = " + pc);
    */
    this._internID(bkpt_id)[tag] = null;
};

JSDBGP.BreakpointScriptMediator.prototype.reportAddSrcBkpt = function(tag, bkpt_id, pc) {
    this._reportBkptAdd(tag, bkpt_id, pc, 'src');
    







};

JSDBGP.BreakpointScriptMediator.prototype.reportAddPPBkpt = function(tag, bkpt_id, pc) {
    this._reportBkptAdd(tag, bkpt_id, pc, 'pp');
    







};

JSDBGP.BreakpointScriptMediator.prototype.reportBkptDelete = function(bkpt_id) {
    var tag_list = this.breakpointsByID[bkpt_id];
    if (!tag_list) return;
    var types = ['pp', 'src'];
    for (var tag in tag_list) {
        var size = 0;
        var hd = this.boundBptsByTag[tag];
        if (!hd) {
            continue;
        }
        for (var i in types) {
            var type = types[i];
            var tagList = hd[type];
            if (!tagList) {
                size = 99;
                continue;
            }
            var pc = tagList.by_id[bkpt_id];
            if (pc) {
                delete tagList.by_id[bkpt_id];
                delete tagList.by_pc[pc];
            }
            size += this.mainObj.utils.keysLength(hd[type].by_pc);
            size += this.mainObj.utils.keysLength(hd[type].by_id);
        }
        if (size == 0) {
            delete this.boundBptsByTag[tag];
        }
    }
    if (bkpt_id in this.breakpointsByID) {
        delete this.breakpointsByID[bkpt_id][tag];
        if (this.mainObj.utils.keysLength(this.breakpointsByID[bkpt_id]) == 0) {
            delete this.breakpointsByID[bkpt_id];
        }
    }







};

JSDBGP.BreakpointScriptMediator.prototype.doScriptDeleted = function(tag) {
    var hd = this.boundBptsByTag[tag];
    if (!hd) {
        return;
    }
    var types = ['pp', 'src'];
    for (var i in types) {
        var type = types[i];
        var tagList = hd[type];
        for (var bkpt_id in tagList.by_id) {
            if (this.breakpointsByID[bkpt_id]) {
                delete this.breakpointsByID[bkpt_id][tag];
            }
            var pc = tagList.by_id[bkpt_id];
            if (pc) {
                this.mainObj.breakpointManager.removeBreakpointByTag(tag, bkpt_id);
                delete tagList.by_id[bkpt_id];
                delete tagList.by_pc[pc];
            }
            if (type == 'pp') {
                this.mainObj.breakpointManager.notifyBreakpointRemove(bkpt_id);
            }
        }
    }
    delete this.boundBptsByTag[tag];
    






};

JSDBGP.BreakpointScriptMediator.prototype.getBoundTags = function(bkpt_id) {
    var tags = [];
    if (bkpt_id in this.breakpointsByID) {
        for (var tag in this.breakpointsByID[bkpt_id]) {
            tags.push(tag);
        }
    }
    return tags;
};

// => { pc => bkpt_id }
JSDBGP.BreakpointScriptMediator.prototype.getBoundSrcBkpts = function(tag) {
    var hd = this.boundBptsByTag[tag];
    if (!hd) {
        return null;
    }
    return hd.src ? hd.src.by_pc : null;
};

// => { pc => bkpt_id }
JSDBGP.BreakpointScriptMediator.prototype.getBoundPPBkpts = function(tag) {
    var hd = this.boundBptsByTag[tag];
    if (!hd) {
        return null;
    }
    return hd.pp ? hd.pp.by_pc : null;
};

JSDBGP.BreakpointScriptMediator.prototype.getAnyBkpt = function(tag, pc) {
    var hd = this.boundBptsByTag[tag];
    if (!hd) {
        return null;
    }
    var types = ['pp', 'src'];
    for (var i in types) {
        var type = types[i];
        var tagList = hd[type];
        if (pc in tagList.by_pc) {
            return tagList.by_pc[pc];
        }
    }
    return null;
};

JSDBGP.BreakpointScriptMediator.prototype.getPC = function(tag, bkpt_id) {
    var hd = this.boundBptsByTag[tag];
    if (!hd) {
        return null;
    }
    if (bkpt_id in hd.pp.by_id) {
        return hd.pp.by_id[bkpt_id];
    } else if (bkpt_id in hd.src.by_id) {
        return hd.src.by_id[bkpt_id];
    } else {
        return null;
    }
};

// Manage the list of locations of breakpoints

JSDBGP.BreakpointScriptMediator.UnboundBreakpointLocations = function(mainObj) {
    this.sourceBreakpoints = {}; // url => {lineNo => {id => null}}, url could be dbgp
    this.mainObj = mainObj;
    this.log = JSDBGP.Logging.getLogger("winger.breakpointScriptMediator.unboundBreakpointLocations");
    this.log.setLevel(JSDBGP.Logging.DEBUG);
};

JSDBGP.BreakpointScriptMediator.UnboundBreakpointLocations.prototype.addLoc = function(url, lineno, bkpt_id) {
    if (!(url in this.sourceBreakpoints)) {



        this.sourceBreakpoints[url] = {};
    }
    var h = this.sourceBreakpoints[url];
    if (!(lineno in h)) {



        h[lineno] = {};
    }
    h[lineno][bkpt_id] = null;
    




};

JSDBGP.BreakpointScriptMediator.UnboundBreakpointLocations.prototype.deleteLoc = function(url, lineno, bkpt_id) {
    if (!(url in this.sourceBreakpoints)) {



        return;
    }
    var h = this.sourceBreakpoints[url];
    if (!(lineno in h)) {



        return;
    }
    delete h[lineno][bkpt_id];
    if (this.mainObj.utils.keysLength(h[lineno]) == 0) {



        delete h[lineno];
        if (this.mainObj.utils.keysLength(this.sourceBreakpoints[url]) == 0) {



            delete this.sourceBreakpoints[url];
        }
    }
    




};

JSDBGP.BreakpointScriptMediator.UnboundBreakpointLocations.prototype.getLoc = function(url) {
    return url in this.sourceBreakpoints ? this.sourceBreakpoints[url] : null;
};

JSDBGP.BreakpointScriptMediator.UnboundBreakpointLocations.prototype.getURLs = function() {
    var urls = [];
    for (var url in this.sourceBreakpoints) {
        urls.push(url);
    }
    return urls;
};
